(*
This file is part of Ceve.

Ceve is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Ceve is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Ceve; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*)

(**
Writer for the EGraph format.
@author Jaap Boender
*)

open Egraph
open Metadata

{{ namespace "http://graphml.graphdrawing.org/xmlns/1.0rc" }}
{{ namespace edos = "http://www.edos-project.org/egraph" }}

let node_from_package (pkg: pkg_metadata): {{ node_type }} =
begin
	{{
		<node package={: pkg.name :} id={: pkg.name :}>[
			<desc>{: pkg.summary :}
			<data key="nodedata">[
				<edos:version number={: pkg.version :} filename={: pkg.filename :}>[]
			]	
		]
	}}
end;;

let edges_from_packages (pkgs: pkg_metadata list): {{ edge_type | hyperedge_type }} list =
let operator_of vs =
begin
	match vs with
	| AnyVersion -> "all"
	| Less _ -> "lt"
	| LessEqual _ -> "le"
	| Equal _ -> "eq"
	| GreaterEqual _ -> "ge"
	| Greater _ -> "gt"
end
and target_of vs =
begin
	match vs with
	| AnyVersion -> ""
	| Less v -> v
	| LessEqual v -> v
	| Equal v -> v
	| GreaterEqual v -> v
	| Greater v -> v
end
and source_ht = Hashtbl.create (List.length pkgs) in
let result = ref [] in
begin
	List.iter (fun pkg ->
	(* Depends: *)
	let run_ht = try Hashtbl.find source_ht (pkg.name, "run")
	with Not_found -> Hashtbl.create (List.length pkg.depends) in
		List.iter (fun disj ->
			let targets = List.map (fun x -> x.target) disj in
			let version_list = try Hashtbl.find run_ht targets
			with Not_found -> List.map (fun x -> []) disj in
				Hashtbl.replace run_ht targets (List.map2 (fun old add ->
					add::old	
				) version_list (List.map (fun x -> (pkg.version, x.version_spec)) disj))
		) pkg.depends;
		Hashtbl.replace source_ht (pkg.name, "run") run_ht;
	(* Pre-Depends: *)
	let install_ht = try Hashtbl.find source_ht (pkg.name, "install")
	with Not_found -> Hashtbl.create (List.length pkg.pre_depends) in
		List.iter (fun disj ->
			let targets = List.map (fun x -> x.target) disj in
			let version_list = try Hashtbl.find install_ht targets
			with Not_found -> List.map (fun x -> []) disj in
				Hashtbl.replace install_ht targets (List.map2 (fun old add ->
					add::old	
				) version_list (List.map (fun x -> (pkg.version, x.version_spec)) disj))
		) pkg.pre_depends;
		Hashtbl.replace source_ht (pkg.name, "install") install_ht;
	(* Recommends: *)
	let recommend_ht = try Hashtbl.find source_ht (pkg.name, "recommend")
	with Not_found -> Hashtbl.create (List.length pkg.recommends) in
		List.iter (fun disj ->
			let targets = List.map (fun x -> x.target) disj in
			let version_list = try Hashtbl.find recommend_ht targets
			with Not_found -> List.map (fun x -> []) disj in
				Hashtbl.replace recommend_ht targets (List.map2 (fun old add ->
					add::old	
				) version_list (List.map (fun x -> (pkg.version, x.version_spec)) disj))
		) pkg.recommends;
		Hashtbl.replace source_ht (pkg.name, "recommend") recommend_ht;
	(* Suggests: *)
	let suggest_ht = try Hashtbl.find source_ht (pkg.name, "suggest")
	with Not_found -> Hashtbl.create (List.length pkg.suggests) in
		List.iter (fun disj ->
			let targets = List.map (fun x -> x.target) disj in
			let version_list = try Hashtbl.find suggest_ht targets
			with Not_found -> List.map (fun x -> []) disj in
				Hashtbl.replace suggest_ht targets (List.map2 (fun old add ->
					add::old	
				) version_list (List.map (fun x -> (pkg.version, x.version_spec)) disj))
		) pkg.suggests;
		Hashtbl.replace source_ht (pkg.name, "suggest") suggest_ht;
	(* Enhances: *)
	let enhance_ht = try Hashtbl.find source_ht (pkg.name, "enhance")
	with Not_found -> Hashtbl.create (List.length pkg.enhances) in
		List.iter (fun disj ->
			let targets = List.map (fun x -> x.target) disj in
			let version_list = try Hashtbl.find enhance_ht targets
			with Not_found -> List.map (fun x -> []) disj in
				Hashtbl.replace enhance_ht targets (List.map2 (fun old add ->
					add::old	
				) version_list (List.map (fun x -> (pkg.version, x.version_spec)) disj))
		) pkg.enhances;
		Hashtbl.replace source_ht (pkg.name, "enhance") enhance_ht;
	(* Conflicts: *)
	let conflict_ht = try Hashtbl.find source_ht (pkg.name, "conflict")
	with Not_found -> Hashtbl.create (List.length pkg.conflicts) in
		List.iter (fun dep ->
			let targets = [dep.target] in
			let version_list = try Hashtbl.find conflict_ht targets
			with Not_found -> [[]] in
				Hashtbl.replace conflict_ht targets (List.map2 (fun old add ->
					add::old
				) version_list [(pkg.version, dep.version_spec)])
		) pkg.conflicts;
		Hashtbl.replace source_ht (pkg.name, "conflict") conflict_ht;
	(* Replaces: *)
	let replace_ht = try Hashtbl.find source_ht (pkg.name, "replace")
	with Not_found -> Hashtbl.create (List.length pkg.replaces) in
		List.iter (fun dep ->
			let targets = [dep.target] in
			let version_list = try Hashtbl.find replace_ht targets
			with Not_found -> [[]] in
				Hashtbl.replace replace_ht targets (List.map2 (fun old add ->
					add::old
				) version_list [(pkg.version, dep.version_spec)])
		) pkg.replaces;
		Hashtbl.replace source_ht (pkg.name, "replace") replace_ht;
	(* Provides: *)
	let provide_ht = try Hashtbl.find source_ht (pkg.name, "provide")
	with Not_found -> Hashtbl.create (List.length pkg.provides) in
		List.iter (fun dep ->
			let targets = [dep.target] in
			let version_list = try Hashtbl.find provide_ht targets
			with Not_found -> [[]] in
				Hashtbl.replace provide_ht targets (List.map2 (fun old add ->
					add::old
				) version_list [(pkg.version, dep.version_spec)])
		) pkg.provides;
		Hashtbl.replace source_ht (pkg.name, "provide") provide_ht
	) pkgs;
	Hashtbl.iter (fun (sn, tp) tns_versions ->
		Hashtbl.iter (fun tns versions ->
			if List.length tns = 1 then
				result := {{ <edge type={: tp :} source={: sn :} target={: List.hd tns :}>[
					<data key="edgedata">		
						{: List.map (fun (sv, tv) -> 
								{{ <edos:version operator={: operator_of tv :} source={: sv :} target={: target_of tv :}>[] }}
							) (List.hd versions)
						:}
				]}}::!result
			else
				result := {{ <hyperedge type={: tp :}>(
					[<endpoint type="out" node={: sn :}>[]] @
					{: List.map2 (fun tn stvs -> 
						{{ <endpoint type="in" node={: tn :}>[
								<data key="endpointdata">(
									{: List.map (fun (sv, tv) -> {{ <edos:version operator={: operator_of tv :} source={: sv :} target={: target_of tv :}>[] }}) stvs :} 
						)]}}
					) tns versions :}
				)}}::!result
		) tns_versions
	) source_ht;
	!result
end;;

let output_egraph (pkgs: pkg_metadata list): unit =
let vt_name (vt: Options.version_types): string =
begin
	match vt with
		Options.Unspecified -> "unspecified"
	| Options.RPMVersions -> "rpm"
	| Options.DebianVersions -> "debian"
end
and packages = List.rev_map node_from_package pkgs
and edges = edges_from_packages pkgs in
begin
	prerr_endline "[XML] Outputting EGraph file...";
	let (output: {{ graphml_type }}) = {{
		<graphml>[
			<graph edgedefault="directed" version_type={: vt_name !Options.version_type :}>(
				{: packages :} @
				{: edges :}
			)
		]
	}} in
	Ocamlduce.Print.print_xml (output_string !Options.output_channel) output
end;;
