(*
This file is part of Ceve.

Ceve is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Ceve is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Ceve; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*)

(**
Writer for the Graphviz (DOT) format.
@author Jaap Boender
*)

open Metadata

let output_node (oc: out_channel) (pkg: pkg_metadata): unit =
begin
	output_string oc ("\"" ^ pkg.name ^ "=" ^ pkg.version ^ "\" [shape=record, color=black, label=\"{" ^ pkg.name ^ "|" ^ pkg.version ^ "}\"];\n")
end;;

let output_provides (oc: out_channel) (pkg: pkg_metadata): unit =
begin
	List.iter (fun pv ->
		output_string oc ("\"" ^ pv.target ^ "\" -> \"" ^ pkg.name ^ "=" ^ pkg.version ^ "\" [style=dashed];\n")
	) pkg.provides
end;;

let get_depname (ht: Dependencies.pkg_hashtable) (dep: pkg_dependency): string =
begin
	let res = String.concat "/" 
	(let versions = Hashtbl.find_all ht dep.target in
		List.flatten (List.map (fun (v, md) ->
			if Dependencies.can_provide dep.version_spec (Equal v) then 
				[md.name ^ "=" ^ md.version]
			else
				[]
		) versions)) in
		if res = "" then dep.target else res
end

let get_disjname (ht: Dependencies.pkg_hashtable) (disj: pkg_dependency list): string =
begin
	String.concat "/" (List.map (get_depname ht) disj)
end;;

let output_disjunctions (oc: out_channel) (ht: Dependencies.pkg_hashtable) (pkg: pkg_metadata): unit =
let output_disj (disj: pkg_dependency list) =
begin
	if List.length disj > 1 then
	let disjname = get_disjname ht disj in
		output_string oc ("\"" ^ disjname ^ "\" [shape=point, color=black];\n");
		List.iter (fun dep ->
			output_string oc ("\"" ^ disjname ^ "\" -> \"" ^ (get_depname ht dep)  ^ "\" [style=dotted];\n")
		) disj
end in
begin
	List.iter output_disj pkg.depends;
	List.iter output_disj pkg.pre_depends;
end;;

let output_dependencies (oc: out_channel) (ht: Dependencies.pkg_hashtable) (pkg: pkg_metadata): unit =
let output_with_disjs (e: pkg_dep_expression) (style: string) =
begin
	List.iter (fun disj ->
		if List.length disj > 1 then
		let disjname = get_disjname ht disj in
			output_string oc ("\"" ^ pkg.name ^ "=" ^ pkg.version ^ "\" -> \"" ^ disjname ^ "\" [" ^ style ^ "];\n")
		else
			List.iter (fun dep ->
				output_string oc ("\"" ^ pkg.name ^ "=" ^ pkg.version ^ "\" -> \"" ^ (get_depname ht dep) ^ "\" [" ^ style ^ "];\n")
			) disj
	) e
end
and output_without_disjs (e: pkg_dependency list) (style: string) =
begin
	List.iter (fun dep ->
		output_string oc ("\"" ^ pkg.name ^ "=" ^ pkg.version ^ "\" -> \"" ^ (get_depname ht dep) ^ "\" [" ^ style ^ "];\n")
	) e
end in
begin
	output_with_disjs pkg.depends "color=black";
	output_with_disjs pkg.pre_depends "color=blue";
	output_without_disjs pkg.conflicts "color=red";
	output_without_disjs pkg.replaces "color=orange";
	output_with_disjs pkg.enhances "color=green";
	output_with_disjs pkg.recommends "color=yellow";
	output_with_disjs pkg.suggests "color=yellow, style=dashed"
end;;

let output_graphviz (pkgs: pkg_metadata list): unit =
begin
	prerr_endline "[Graphviz] Outputting Graphviz file...";
	let ht = Dependencies.create_pkgs_hashtable pkgs in
	output_string !Options.output_channel "digraph G {\n";
	output_string !Options.output_channel "node [shape=ellipse, color=red];\n";
	List.iter (output_node !Options.output_channel) pkgs;
	List.iter (output_provides !Options.output_channel) pkgs;
	List.iter (output_disjunctions !Options.output_channel ht) pkgs;
	List.iter (output_dependencies !Options.output_channel ht) pkgs;
	output_string !Options.output_channel "}\n"
end;;
